function[Lopt, Ropt, zopt, Sopt, infos, options] = fixedrank_hankelmatrixapproximation(problem, Uinit, options)
    % Problem description
    d = problem.d;
    T = problem.T;
    r = problem.r;
    C = problem.C;
    y = problem.y;
    
    if isfield(problem,'y_test') && ~isempty(problem.y_test)
        y_test = problem.y_test;
    end
    
    % Local defaults for options
    localdefaults.maxiter = 50; % Max iterations.
    localdefaults.tolgradnorm = 1e-6; % Absolute tolerance on Gradnorm.
    localdefaults.maxinner = 30; % Max inner iterations for the tCG step.
    localdefaults.tolrelgradnorm = 1e-10; % Gradnorm/initGradnorm tolerance.
    localdefaults.method = 'tr'; % Default solver is trustregions (TR).
    localdefaults.innerpcgdottol = 1e-10;
    localdefaults.innerpcgdotmaxiter = 500;
    localdefaults.innerpcgtol = 1e-8;
    localdefaults.innerpcgmaxiter = 500;
    localdefaults.reg = 0; % If non zero, then W = L*R + reg*S.
    localdefaults.preconreg = 1e-5; % Regularization for preconditioner.
    localdefaults.verbosity = 2; % Default is print all the details.
    
    % Compute linear indices for the Hankel structure for size d+1-by-T+1.
    hankelidx = hankel(1:d+1,d+1:(T+1-1)+d+1);        
    hankelnormalizer = [1:d+1, (d+1)*ones(1,d +1 + T+1 - 1 - 2*(d+1)), (d+1):-1:1]'; % Only for fat matrices, will be different for long matrices
    if d+1 < T+1
        % Do nothing
    elseif d+1 > T+1
        hankelidx = hankelidx';
    else
        hankelnormalizer = [1:d+1, (d+1)*ones(1,d+1 + T+1 - 1 - 2*(d+1)), (d + 1-1):-1:1]';
    end
    localdefaults.hankelidx = hankelidx;
    localdefaults.hankelnormalizer = hankelnormalizer;
    
    % Manifold geometry
    problem.M = spectrahedronfactory(d, r);
    
    %     % For Sanity check use this factory.
    %     problem.M = euclideanfactory(d, r);
    
    
    % Initialization
    if ~exist('Uinit', 'var')
        Uinit = [];
    end
    
    if ~exist('options', 'var') || isempty(options)
        options = struct();
    end
    options = mergeOptions(localdefaults, options);
    
    
    zold = zeros(d + T - 1, 1);
    Sold = zeros(d,T);
    eyer = eye(r);
    
    % Cost
    problem.cost = @cost;
    function [f, store] = cost(U, store)
        if ~isfield(store, 'z')
            % Compute preconditioner
            UtU = U'*U;
            L = (eyer  + (UtU)/options.preconreg);
            store.L = L;
            
            % Compute z and S.
            [store.z, store.S] = HankelsquareLoss_dual_get_zS(U, C, y, options, L, zold, Sold);
            
            %             % Sanity checks.
            %             % Optimization approach to solve z and S
            %             [z1, S1] = HankelsquareLoss_dual_get_zS_opt_BM(U, C, y, [], options,  zold, Sold);
            %             norm(z1 - store.z)
            %             norm(S1 - store.S, 'fro')
            
            store.UtS = U'*store.S;
        end
        z = store.z;
        S = store.S;
        UtS = store.UtS;
        f = (y'*z) -  (z'*z)/(4*C) - 0.5*(norm(UtS, 'fro')^2) ...
            - (0.5*options.reg)*(norm(S, 'fro')^2);
        %         disp(options.reg);
    end
    
    % Gradient
    problem.egrad = @egrad;
    function [grad, store] = egrad(U, store)
        if ~isfield(store, 'z')
            [~, store] = cost(U, store);
        end
        S = store.S;
        UtS = store.UtS;
        grad = -S*((UtS)');
    end
    
    
    % Hessian
    problem.ehess = @ehess;
    function[graddot, store] = ehess(U, Udot, store)
        if ~isfield(store, 'z')
            [~, store] = cost(U, store);
        end
        z = store.z;
        S = store.S;
        UtS = store.UtS;
        
        [zdot, Sdot] = HankelsquareLoss_dual_get_zdotSdot(U, Udot, C, y, z, S, UtS, options, store.L, zold, Sold);
        
        UtSdot = U'*Sdot;
        UdottS = Udot'*S;
        
        graddot =  - Sdot*(UtS)' - S*(UdottS + UtSdot)';
    end
    
    
    % Stats that we compute every iteration
    options.statsfun = @mystatsfun;
    function [stats, store] = mystatsfun(problem, U, stats, store)
        z = store.z;
        S = store.S;
        
        primal_cost = stats.cost; % Equal to the partial dual cost.
        sigmax_S = svds(S, 1); % Maximum singular value of S.
        
        dual_cost = (y'*z) - (z'*z)/(4*C) - 0.5*sigmax_S^2 - (0.5*options.reg)*(norm(S, 'fro')^2);
        stats.duality_gap = abs(primal_cost - dual_cost); % Absolute duality gap
        stats.rel_duality_gap = abs(stats.duality_gap/dual_cost); % Relative duality gap
        
        % Primal low-rank factors
        L = U;
        R = store.UtS;
        W = L*R + options.reg*S;
        w = [W(1:d,1); (W(end,2:end))'];
        
        
        % Compute train RMSE
        stats.loss =  sum(sum(((w - y).^2))); % Square error on train data.
        stats.loss_rmse = sqrt(stats.loss/(d + T -1)); % RMSE on train data.
        
        if exist('y_test', 'var')
            stats.loss_test = sum(sum(((w-y_test).^2))); % Square error on test data.
            stats.loss_test_rmse = sqrt(stats.loss_test/(d + T -1)); % RMSE on test data.
            
        end
    end
    
    % Additional stopping criteria
    options.stopfun = @mystopfun;
    function stopnow = mystopfun(problem, U, info, last)
        stopnow = (last >= 3 && info(last).gradnorm/info(1).gradnorm < options.tolrelgradnorm);
    end
    
    %     % Gradient and Hessian checks
    %     checkgradient(problem);
    %     pause;
    %     checkhessian(problem);
    %     pause;
    
    % Solver
    if strcmpi('TR', options.method)
        % Riemannian trustregions
        [Uopt,~,infos] = trustregions(problem, Uinit, options);
    elseif strcmpi('CG', options.method)
        % Riemannian conjugategradients
        options.beta_type = 'H-S';
        options.linesearch = @linesearch;
        options.ls_contraction_factor = .2;
        options.ls_optimism = 1.1;
        options.ls_suff_decr = 1e-4;
        options.ls_max_steps = 25;
        [Uopt,~,infos] = conjugategradient(problem, Uinit, options);
    end
    
    % Compute optimal z and S
    [zopt, Sopt] = HankelsquareLoss_dual_get_zS(Uopt, C, y, options);
    
    % Optimal low-rank factors
    Lopt = Uopt;
    Ropt = Uopt'*(Sopt);
    
end