/*=================================================================
% function R = spbuildmatrix(Ct, U, lambda)
%
% Specific function to reduce the computation time in a bottleneck
% portion of the Matlab code for our low-rank matrix completion
% software.
%
% Ct is a *sparse* m-by-n real matrix, U is a (real)
% m-by-r matrix and lambda is a strictly positive real number.
% R is a cell containing n upper triangular r-by-r matrices: the
% Cholesky factors of the diagonal blocks of A.'*A. See the notes
% accompanying the software for details.
%
% The equivalent Matlab code is the following:
%
%         AtA = cell(n, 1);
%         for i = 1 : n
%             
%             I = find(Ct(:, i));
%             scaledUi = U(I, :);
%             AtAi = scaledUi.'*scaledUi + eye(r) / (2*lambda);
%             
%             % We actually store the Cholesky factorization
%             AtA{i} = chol(AtAi);
%             
%         end
%
% 
% Compile with: mex -lmwlapack -lmwblas -largeArrayDims myspbuildmatrix_parallel.c CFLAGS="\$CFLAGS -fopenmp" LDFLAGS="\$LDFLAGS -fopenmp"
%
% Code originally written by Nicolas Boumal <nicolasboumal@gmail.com>.
% Original code has been modified by Pratik Jawanpuria <pratik.iitb@gmail.com> and Bamdev Mishra <bamdevm@gmail.com>.
 *=================================================================*/

#include "mex.h"
#include "matrix.h"
#include "lapack.h"
#include "blas.h"
#include <math.h>
#include <sys/time.h>
#include <omp.h>


/* Timing functions */
struct timeval start()
{
    struct timeval tm ;
    gettimeofday(&tm, NULL);
    return tm ;
}

void stop(struct timeval tm1)
{
    struct timeval tm2;
    gettimeofday(&tm2, NULL);
    unsigned long long t = 1000 * (tm2.tv_sec - tm1.tv_sec) + (tm2.tv_usec - tm1.tv_usec) / 1000;
    printf("%llu ms\n", t);
}

unsigned long long stopAndGet(struct timeval tm1)
{
    struct timeval tm2;
    gettimeofday(&tm2, NULL);
    unsigned long long t = 1000 * (tm2.tv_sec - tm1.tv_sec) + (tm2.tv_usec - tm1.tv_usec) / 1000;
    return t ;
}

double stopAndGetDouble(struct timeval tm1)
{
    return (double)stopAndGet(tm1) ;
}

unsigned long long getMeanTimes(unsigned long long* times, int* nOfIterations, int maxthreads)
{
    unsigned long long meanTime = 0 ;
    int i = 0 ;
    for(i = 0 ; i < maxthreads ; i++)
    {
        if(nOfIterations[i] > 0)
            meanTime += times[i] ; /* /nOfIterations[i] ;   */
    }
    return meanTime ;
}


/* Input Arguments */
#define	Ct	    prhs[0]
#define	Ut	    prhs[1]
#define	lambda	prhs[2]
#define numThreads prhs[3]

/* Output Arguments */
#define R plhs[0]
void mexFunction(
          int nlhs, mxArray *plhs[], 
		  int nrhs, const mxArray* prhs[])
{
    mwIndex ii;
    mwSize m, n, r;
    double *Ctvals, *Utvals;
    char *TRANS = "T";
    char *NOTRANS = "N";
    double lambdaval;
    mwIndex *CtIr, *CtJc;
    
    if(nrhs != 4 || nlhs != 1)
        mexErrMsgTxt("Invalid number of input/output parameter. Need 4 inputs and 1 output.");
    
    m = mxGetM(Ct);
    n = mxGetN(Ct);
    r = mxGetM(Ut);
    if(mxGetN(Ut) != m)
        mexErrMsgTxt("Dimensions mismatch in input arguments.");
    
    if(!mxIsSparse(Ct))
        mexErrMsgTxt("Ct ought to be a sparse matrix.");
    
    lambdaval = 1.0 / (2.0 * mxGetScalar(lambda)) ;
    
    CtIr = mxGetIr(Ct);
    CtJc = mxGetJc(Ct);
    Ctvals = mxGetPr(Ct);
    Utvals = mxGetPr(Ut);
    
    int maxNrThreads = (int) mxGetScalar(numThreads);
    
    /* Create a suitable cell */
    R = mxCreateCellMatrix(n, 1);
    for(ii = 0; ii < n; ++ii)
    {
        /* allocate space for an r-by-r matrix in each cell entry */
        mxSetCell(R, ii, mxCreateDoubleMatrix(r, r, mxREAL));
    }

    /* Create a dummy cell with matrices which will be used for memory allocation*/
    mxArray *scaledUt;
    scaledUt = mxCreateCellMatrix(maxNrThreads, 1);
    for(ii = 0; ii < maxNrThreads; ++ii)
    {
        /* allocate space for an r-by-r matrix in each cell entry */
        mxSetCell(scaledUt, ii, mxCreateDoubleMatrix(r, m, mxREAL));
    }

    omp_set_num_threads(maxNrThreads);
    #pragma omp parallel
    {
        mwIndex i;
        int id = omp_get_thread_num();
        int nthrds = omp_get_num_threads();
        mxArray *scaledUti = mxGetCell(scaledUt, id);
        double *scaledUtivals = mxGetPr(scaledUti);
        char *uplo = "U";
        /*ptrdiff_t info = 0;*/
        mxArray *Ri;
        double *Rivals;
        mwIndex *Ctcol;
        mwSize nb;
        ptrdiff_t NB;
        int j, k, temp1, temp2, temp3;
        ptrdiff_t N, M, RR;
        double one = 1.0;
        double zero = 0.0;
        N  = (ptrdiff_t) n;
        M  = (ptrdiff_t) m;
        RR = (ptrdiff_t) r;
        /*#pragma omp for private(info,uplo,Ri,Rivals,Ctcol,nb,NB)*/
        for(i = id; i < n; i+=nthrds)
        {
            /*printf("i: %d, id: %d, nthrds: %d\n",i,id,nthrds);*/
            ptrdiff_t info = 0;
            Ri = mxGetCell(R, i);
            Rivals = mxGetPr(Ri);
            
            /* compute the entries of the matrix R{i}
               nb is the number of nonzero entries in the i-th column of Ct
               Ctcol is a pointer to an array of nb elements: the indices of
               rows where a nonzero element exists on column i of Ct. */
            nb = CtJc[i+1]-CtJc[i];
            /* mexPrintf("%d ", nb); */
            Ctcol = CtIr+CtJc[i];
            if(nb > 0)
            {
                for(j = 0; j < nb; ++j)
                {
                    temp1 = j*r;
                    temp2 = Ctcol[j]*r;
                    for(k = 0; k < r; ++k)
                    {
                        /* scaledUi(j, k) =  U(Ctcol[j], k) */
                        /*scaledUivals[j+k*m] = Uvals[Ctcol[j]+k*m];*/
                        scaledUtivals[k+temp1] = Utvals[k+temp2];
                    }
                }
                /*   mexPrintf("\n"); */
                /* compute the matrix product of the interesting part of
                   scaledUi with itself (transposed): Ri = scaledUi.'*scaledUi */
                NB = (ptrdiff_t) nb;
                dgemm(NOTRANS, TRANS, &RR, &RR, &NB, &one,
                      scaledUtivals, &RR, scaledUtivals, &RR, &zero, Rivals, &RR);
                /*dgemm(TRANS, NOTRANS, &RR, &RR, &NB, &one,
                      scaledUivals, &M, scaledUivals, &M, &zero, Rivals, &RR);*/
            }
            else
            {
                for(j = 0; j < r*r; ++j)
                    Rivals[j] = 0.0;
            }
            /* add lambda*eye(r) to Ri */
            for(j = 0; j < r; ++j)
                Rivals[j*(r+1)] += lambdaval;
            
            /* remplace R{i} by its Cholesky factor by calling Lapack's dpotrf. */
            dpotrf(uplo, &RR, Rivals, &RR, &info);
            /* write zeroes on the lower triangular parts of Ri */
            /*for(j = 1; j < r; ++j)
                for(k = 0; k < j; ++k)
                    Rivals1[j+r*k] = 0.0;*/
            for(k = 0; k < r-1; ++k)
            {
                temp3 = r*k;
                for(j = k+1; j < r; ++j)
                    Rivals[j+temp3] = 0.0;
            }
            if(info < 0)
            {
                /*mexErrMsgTxt("dpotrf (in spbuildmatrix): invalid argument.");*/
                printf("ErrMsgTxt: dpotrf (in spbuildmatrix): invalid argument.");
            }
            else if(info > 0)
            {
                /*mexErrMsgTxt("dpotrf (in spbuildmatrix): a leading minor is not positive definite.");*/
                printf("dpotrf (in spbuildmatrix): a leading minor is not positive definite.");
            }
        }
    }
    /*mexPrintf("Time dpotrf: ");
    stop(start_forW2);*/

    mxDestroyArray(scaledUt);
    
    return;
}


