% This sample demo test code creates a synthetic matrix completion data set
% of dimensions dxT and call both our first- and second- order algorithms,
% Proposed-cg-sq and Proposed-tr-sq respectively, to complete the matrix.

clear; clc; close all;
randomize_seed = 25;
rng(randomize_seed);
dataset = 'synthetic';

sp = 1;
fprintf('Dataset: %s, Split: %d of %d\n', dataset, sp, sp);
fprintf('randomize_seed: %d\n', randomize_seed);

% A synthetic problem instance.
d = 500; % Config used in paper d = 5000
T = 50000; % Config used in paper T = 500000
truerank = 5;
OS = 6; % Over sampling. No of known entries relative to dimension. Typically 4,5,8.
noiseFac = 0; 1e-6; % Noise factor % good config: 0
fprintf('d: %d, T: %d, truerank: %d, OS: %d, noiseFac: %e\n',d,T,truerank,OS,noiseFac);

% Call function to create synthetic data set
[I, J, Y, I_test, J_test, Y_test] = createSyntheticData(d, T, truerank, OS, noiseFac);
fprintf('Number of train instances: %d\n',length(Y));
fprintf('Number of test instances: %d\n',length(Y_test));
%% Create Problem structure and set hyper-parameters
% Set regularization parameter
C = 1e7;
fprintf('C: %e\n',C);
% Rank that we impose
rank = truerank;
fprintf('rank: %d\n',rank);

% Problem structure
problem.d = d;
problem.T = T;
problem.r = rank;
problem.C = C;
problem.I = I;
problem.J = J;
problem.Y = Y;
problem.I_test = I_test;
problem.J_test = J_test;
problem.Y_test = Y_test;

%% Proposed-tr-sq algorithm call

% Options are not mandatory
opts.maxiter = 20;
opts.tolgradnorm = 1e-5;% 1e-5
opts.maxinner = 30;
opts.tolrelgradnorm = 1e-20;% 1e-20
opts.method='tr';
opts.verbosity = 2; % 2 for full output.
opts

fprintf('------------------ Proposed-tr-sq ------------------\n')

initproposedtrstart = tic;
% Initialization
mysparsemat = sparse(double(problem.I), double(problem.J), ones(length(problem.I), 1), problem.d, problem.T, length(problem.I));
mysetsparseentries(mysparsemat , problem.Y);
[Uinit, ~, ~]= svds(mysparsemat, problem.r);
Uinit = Uinit/norm(Uinit, 'fro');
initproposedtrtime = toc(initproposedtrstart);
fprintf('SVD based warm start\n');
fprintf('initproposedtime: %e\n',initproposedtrtime);

[~, ~, ~, infos, options] = fixedrank_matrixcompletion(problem,Uinit,opts);
proposed_plot_rmse = [infos.loss_test_rmse];
proposed_plot_time = [infos.time]+initproposedtrtime;
proposed_plot_dualitygap = [infos.rel_duality_gap];

%% Proposed-cg-sq algorithm call

% Options are not mandatory
optscg.maxiter = 100;
optscg.tolgradnorm = 1e-5;% 1e-5
optscg.tolrelgradnorm = 1e-20;% 1e-20
optscg.method='cg';
optscg

fprintf('------------------ Proposed-cg-sq ------------------\n')

initproposedcgstart = tic;
% Initialization
mysparsemat = sparse(double(problem.I), double(problem.J), ones(length(problem.I), 1), problem.d, problem.T, length(problem.I));
mysetsparseentries(mysparsemat , problem.Y);
[Uinit, ~, ~]= svds(mysparsemat, problem.r);
Uinit = Uinit/norm(Uinit, 'fro');
initproposedcgtime = toc(initproposedcgstart);
fprintf('SVD based warm start\n');
fprintf('initproposedcgtime: %e\n',initproposedcgtime);

[~, ~, ~, infoscg, optionscg] = fixedrank_matrixcompletion(problem,Uinit,optscg);
proposed_cg_plot_rmse = [infoscg.loss_test_rmse];
proposed_cg_plot_time = [infoscg.time]+initproposedcgtime;
proposed_cg_plot_dualitygap = [infoscg.rel_duality_gap];

%% Plot: Test RMSE vs Time
fs = 20;
figure;
loglog(proposed_plot_time,proposed_plot_rmse,'-','Color','r','LineWidth',2, 'MarkerSize',13);
hold on;
loglog(proposed_cg_plot_time,proposed_cg_plot_rmse,'-.','Color','b','LineWidth',2, 'MarkerSize',4);
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,'Time (in seconds)','FontSize',fs);
ylabel(ax1,'Test RMSE','FontSize',fs);
legend('Proposed-tr-sq','Proposed-cg-sq','Location','SouthWest');
legend 'boxoff';
% box off;
grid on;
set(gcf,'color','w');

%% Plot: Relative duality gap vs Iteration
fs = 20;
figure;
semilogy(proposed_plot_dualitygap,'-','Color','r','LineWidth',2, 'MarkerSize',10);
hold on;
semilogy(proposed_cg_plot_dualitygap,'-.','Color','b','LineWidth',2, 'MarkerSize',5);
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,'Iteration','FontSize',fs);
ylabel(ax1,'Relative duality gap','FontSize',fs);
legend('Proposed-tr-sq','Proposed-cg-sq', 'Location', 'SouthEast');
legend 'boxoff';
% box off;
grid on;
set(gcf,'color','w');

%% Plot: Relative duality gap vs Time
% Plot for relative duality gap
fs = 20;
figure;
semilogy(proposed_plot_time,proposed_plot_dualitygap,'-','Color','r','LineWidth',2, 'MarkerSize',5);
hold on;
semilogy(proposed_cg_plot_time,proposed_cg_plot_dualitygap,'-.','Color','b','LineWidth',2, 'MarkerSize',5);
hold off;
ax1 = gca;
set(ax1,'FontSize',fs);
xlabel(ax1,'Time (in seconds)','FontSize',fs);
ylabel(ax1,'Relative duality gap','FontSize',fs);
legend('Proposed-tr-sq','Proposed-cg-sq', 'Location', 'SouthEast');
legend 'boxoff';
% box off;
grid on;
set(gcf,'color','w');

