function[Uopt, XZopt, infos, options] = fixedrank_mtfl(problem, Uinit, options)
    % Problem description
    d = problem.d;
    r = problem.r;
    C = problem.C;
    data_train = problem.data;
    if isfield(problem,'data_test') && ~isempty(problem.data_test)
        data_test = problem.data_test;
    end
    
    % Local defaults for options
    localdefaults.maxiter = 50; % Max iterations.
    localdefaults.tolgradnorm = 1e-6; % Absolute tolerance on Gradnorm.
    localdefaults.maxinner = 30; % Max inner iterations for the tCG step.
    localdefaults.tolrelgradnorm = 1e-10; % Gradnorm/initGradnorm tolerance.
    localdefaults.method = 'TR'; % Default solver is trustregions (TR).
    localdefaults.verbosity = 2; % Default is print all the details.
    
    
    % Manifold geometry
    problem.M = spectrahedronfactory(d, r);
    
    % Initialization
    if ~exist('Uinit', 'var')
        Uinit = [];
    end
    
    if ~exist('options', 'var') || isempty(options)
        options = struct();
    end
    options = mergeOptions(localdefaults, options);
    
    % Cost
    problem.cost = @cost;
    function [f, store] = cost(U, store)
        if ~isfield(store, 'UtXZ')
            [store.XZ, store.UtXZ, store.traceYtZ, store.traceZtZ, store.Q_task] = ...
                mtfl_squareLoss_dual_get_Z(U,C,data_train);
        end
        %         XZ = store.XZ; % d x T
        UtXZ = store.UtXZ; % r x T
        traceYtZ = store.traceYtZ; % 1x1
        traceZtZ = store.traceZtZ; % 1x1
        f = traceYtZ - traceZtZ/(4*C) - 0.5*norm(UtXZ,'fro')^2;
        clear UtXZ
    end
    
    % Gradient
    problem.egrad = @egrad;
    function [grad, store] = egrad(U, store)
        if ~isfield(store, 'UtXZ')
            [~, store] = cost(U, store);
        end
        UtXZ = store.UtXZ;
        XZ = store.XZ;
        grad = -(XZ*UtXZ'); % Equivalent to grad = -(Z*ZtU).
        clear UtXZ XZ
    end
    
    % Hessian
    problem.ehess = @ehess;
    function [graddot, store] = ehess(U, Udot,store)
        if ~isfield(store, 'UtXZ')
            [~, store] = cost(U, store);
        end
        UtXZ = store.UtXZ;
        XZ = store.XZ;
        Q_task = store.Q_task;
        % Get Zdot at U and Udot.
        [XZdot] = mtfl_squareLoss_dual_get_Zot(U,C,data_train,Udot,XZ, UtXZ,Q_task);
        
        temp2 = (XZdot*XZ');
        graddot = - ((temp2 + temp2')*U + (XZ*XZ')*Udot);
        clear UtXZ XZ
    end
    
    % Stats that we compute every iteration
    options.statsfun = @mystatsfun;
    function stats = mystatsfun(problem, U, stats, store)
        primal_cost = stats.cost; % Equal to the partial dual cost.
        XZ = store.XZ;
        traceYtZ = store.traceYtZ;
        traceZtZ = store.traceZtZ;
        sigmax_XZ = svds(XZ, 1); % Maximum singular value of Z.
        dual_cost = traceYtZ - traceZtZ/(4*C) - 0.5*sigmax_XZ^2;
        stats.duality_gap = abs(primal_cost - dual_cost); % Absolute duality gap
        stats.rel_duality_gap = abs(stats.duality_gap/dual_cost); % Relative duality gap
        % Compute train RMSE
        UtXZ = store.UtXZ;
        UUtXZ = U*UtXZ;
        T = size(XZ,2);
        train_rmse_tasks = nan(T,1);
        train_nmse_tasks = nan(T,1);
        for t=1:T
            y_train_pred_task = data_train(t).X*UUtXZ(:,t);
            train_nmse_tasks(t) = calculateErr(y_train_pred_task, data_train(t).y, 'nmse');
            train_rmse_tasks(t) = calculateErr(y_train_pred_task, data_train(t).y, 'rmse');
        end
        stats.loss_train_nmse = mean(train_nmse_tasks);
        stats.loss_train_rmse = mean(train_rmse_tasks);
        if exist('data_test','var')
            test_rmse_tasks = nan(T,1);
            test_nmse_tasks = nan(T,1);
            y_test_pred_cell = cell(T,1);
            for t=1:T
                y_test_pred_task = data_test(t).X*UUtXZ(:,t);
                test_nmse_tasks(t) = calculateErr(y_test_pred_task, data_test(t).y, 'nmse');
                test_rmse_tasks(t) = calculateErr(y_test_pred_task, data_test(t).y, 'rmse');
                y_test_pred_cell{t} = y_test_pred_task;
            end
            stats.loss_test_nmse = mean(test_nmse_tasks);
            stats.loss_test_rmse = mean(test_rmse_tasks);
        end
        if options.verbosity > 1
            fprintf('Train NMSE: %e, Test NMSE: %e, Rel. Duality Gap: %e\n',stats.loss_train_nmse,stats.loss_test_nmse,stats.rel_duality_gap);
        end
    end
    
    % Additional stopping criteria
    options.stopfun = @mystopfun;
    function stopnow = mystopfun(problem, U, info, last)
        stopnow = (last >= 3 && info(last).gradnorm/info(1).gradnorm < options.tolrelgradnorm);
    end
    
    
    %         % Gradient and Hessian checks
    %         checkgradient(problem);
    %         pause;
    %         checkhessian(problem);
    %         pause;
    
    % Solver
    if strcmpi('TR', options.method)
        % Riemannian trustregions
        [Uopt,~,infos] = trustregions(problem, Uinit, options);
    elseif strcmpi('CG', options.method)
        % Riemannian conjugategradients
        options.beta_type = 'H-S';
        options.linesearch = @linesearch;
        options.ls_contraction_factor = .2;
        options.ls_optimism = 1.1;
        options.ls_suff_decr = 1e-4;
        options.ls_max_steps = 25;
        [Uopt,~,infos] = conjugategradient(problem, Uinit, options);
    end
    
    % Compute optimal Z
    [XZopt, ~, ~, ~] = mtfl_squareLoss_dual_get_Z(Uopt,C,data_train);
    
end